@@ -1,3 +1,6 @@
+from __future__ import annotations
+
+
 __copyright__ = "Copyright (C) 2009-2021 Andreas Kloeckner"
 
 __license__ = """
@@ -23,14 +26,20 @@ THE SOFTWARE.
 
 import logging
 import sys
+from dataclasses import dataclass
+from typing import TYPE_CHECKING
 
 import pytest
 
 from pytools import Record
+from pytools.tag import tag_dataclass
+
+
+if TYPE_CHECKING:
+    from collections.abc import Sequence
 
 
 logger = logging.getLogger(__name__)
-from typing import FrozenSet
 
 
 def test_memoize_method_clear():
@@ -50,7 +59,7 @@ def test_memoize_method_clear():
     sc.f()
     assert sc.run_count == 1
 
-    sc.f.clear_cache(sc)  # pylint: disable=no-member
+    sc.f.clear_cache(sc)
 
 
 def test_keyed_memoize_method_with_uncached():
@@ -73,7 +82,7 @@ def test_keyed_memoize_method_with_uncached():
     sc.f(18, 19, z=20)
     assert sc.run_count == 2
 
-    sc.f.clear_cache(sc)  # pylint: disable=no-member
+    sc.f.clear_cache(sc)
 
 
 def test_memoize_in():
@@ -173,7 +182,6 @@ def test_memoize_keyfunc():
 
 
 def test_memoize_frozen() -> None:
-    from dataclasses import dataclass
 
     from pytools import memoize_method
 
@@ -190,8 +198,7 @@ def test_memoize_frozen() -> None:
     c0 = FrozenDataclass(10)
     assert c0.double_value() == 20
 
-    # pylint: disable=no-member
-    c0.double_value.clear_cache(c0)  # type: ignore[attr-defined]
+    c0.double_value.clear_cache(c0)
 
     # }}}
 
@@ -213,8 +220,7 @@ def test_memoize_frozen() -> None:
     c1 = FrozenClass(10)
     assert c1.double_value() == 20
 
-    # pylint: disable=no-member
-    c1.double_value.clear_cache(c1)  # type: ignore[attr-defined]
+    c1.double_value.clear_cache(c1)
 
     # }}}
 
@@ -223,8 +229,10 @@ def test_memoize_frozen() -> None:
 def test_spatial_btree(dims, do_plot=False):
     pytest.importorskip("numpy")
     import numpy as np
+
+    rng = np.random.default_rng()
     nparticles = 2000
-    x = -1 + 2*np.random.rand(dims, nparticles)
+    x = -1 + 2*rng.uniform(size=(dims, nparticles))
     x = np.sign(x)*np.abs(x)**1.9
     x = (1.4 + x) % 2 - 1
 
@@ -360,8 +368,6 @@ def test_eoc():
 
     # {{{ test invalid inputs
 
-    import numpy as np
-
     eoc = EOCRecorder()
 
     # scalar inputs are fine
@@ -404,14 +410,14 @@ class FakeArray:
     def __getitem__(self, idx):
         FakeArray.nopes += 1
         if idx > 10:
-            raise IndexError()
+            raise IndexError
 
 
 def test_make_obj_array_iteration():
     pytest.importorskip("numpy")
 
-    from pytools.obj_array import make_obj_array
-    make_obj_array([FakeArray()])
+    import pytools.obj_array as obj_array
+    obj_array.new_1d([FakeArray()])
 
     assert FakeArray.nopes == 0, FakeArray.nopes
 
@@ -424,9 +430,9 @@ def test_obj_array_vectorize(c=1):
     np = pytest.importorskip("numpy")
     la = pytest.importorskip("numpy.linalg")
 
-    # {{{ functions
+    import pytools.obj_array as obj_array
 
-    import pytools.obj_array as obj
+    # {{{ functions
 
     def add_one(ary):
         assert ary.dtype.char != "O"
@@ -436,12 +442,12 @@ def test_obj_array_vectorize(c=1):
         assert x.dtype.char != "O" and y.dtype.char != "O"
         return x * y + c
 
-    @obj.obj_array_vectorized
+    @obj_array.vectorized
     def vectorized_add_one(ary):
         assert ary.dtype.char != "O"
         return ary + c
 
-    @obj.obj_array_vectorized_n_args
+    @obj_array.vectorized_n_args
     def vectorized_two_add_one(x, y):
         assert x.dtype.char != "O" and y.dtype.char != "O"
         return x * y + c
@@ -454,7 +460,7 @@ def test_obj_array_vectorize(c=1):
             assert ary.dtype.char != "O"
             return ary + self.c
 
-        @obj.obj_array_vectorized_n_args
+        @obj_array.vectorized_n_args
         def vectorized_add(self, ary):
             assert ary.dtype.char != "O"
             return ary + self.c
@@ -466,12 +472,12 @@ def test_obj_array_vectorize(c=1):
     # {{{ check
 
     scalar_ary = np.ones(42, dtype=np.float64)
-    object_ary = obj.make_obj_array([scalar_ary, scalar_ary, scalar_ary])
+    object_ary = obj_array.new_1d([scalar_ary, scalar_ary, scalar_ary])
 
     for func, vectorizer, nargs in [
-            (add_one, obj.obj_array_vectorize, 1),
-            (two_add_one, obj.obj_array_vectorize_n_args, 2),
-            (adder.add, obj.obj_array_vectorize, 1),
+            (add_one, obj_array.vectorize, 1),
+            (two_add_one, obj_array.vectorize_n_args, 2),
+            (adder.add, obj_array.vectorize, 1),
             ]:
         input_ary = [scalar_ary] * nargs
         result = vectorizer(func, *input_ary)
@@ -498,12 +504,19 @@ def test_obj_array_vectorize(c=1):
 # }}}
 
 
-def test_tag():
+def test_tag() -> None:
     from pytools.tag import (
-        NonUniqueTagError, Tag, Taggable, UniqueTag, check_tag_uniqueness)
+        NonUniqueTagError,
+        Tag,
+        Taggable,
+        UniqueTag,
+        check_tag_uniqueness,
+    )
 
     # Need a subclass that defines the copy function in order to test.
+    @tag_dataclass
     class TaggableWithCopy(Taggable):
+        tags: frozenset[Tag]
 
         def _with_new_tags(self, tags):
             return TaggableWithCopy(tags)
@@ -604,7 +617,7 @@ def test_unordered_hash():
     # FIXME: Use randbytes once >=3.9 is OK
     lst = [bytes([random.randrange(256) for _ in range(20)])
             for _ in range(200)]
-    lorig = lst[:]
+    lorig = lst.copy()
     random.shuffle(lst)
 
     from pytools import unordered_hash
@@ -663,8 +676,8 @@ def test_sphere_sampling(sampling, visualize=False):
     triangles = mtri.Triangulation(theta, phi)
 
     ax.plot_trisurf(points[0], points[1], points[2], triangles=triangles.triangles)
-    ax.set_xlim([-radius, radius])
-    ax.set_ylim([-radius, radius])
+    ax.set_xlim((-radius, radius))
+    ax.set_ylim((-radius, radius))
     ax.set_zlim([-radius, radius])
     ax.margins(0.05, 0.05, 0.05)
 
@@ -687,40 +700,6 @@ def test_unique_name_gen_conflicting_ok():
     ung.add_names({"a", "b", "c"}, conflicting_ok=True)
 
 
-def test_ignoredforequalitytag():
-    from pytools.tag import IgnoredForEqualityTag, Tag, Taggable
-
-    # Need a subclass that defines _with_new_tags in order to test.
-    class TaggableWithNewTags(Taggable):
-
-        def _with_new_tags(self, tags: FrozenSet[Tag]):
-            return TaggableWithNewTags(tags)
-
-    class Eq1(IgnoredForEqualityTag):
-        pass
-
-    class Eq2(IgnoredForEqualityTag):
-        pass
-
-    class Eq3(Tag):
-        pass
-
-    eq1 = TaggableWithNewTags(frozenset([Eq1()]))
-    eq2 = TaggableWithNewTags(frozenset([Eq2()]))
-    eq12 = TaggableWithNewTags(frozenset([Eq1(), Eq2()]))
-    eq3 = TaggableWithNewTags(frozenset([Eq1(), Eq3()]))
-
-    assert eq1 == eq2 == eq12
-    assert eq1 != eq3
-
-    assert eq1.without_tags(Eq1())
-    with pytest.raises(ValueError):
-        eq3.without_tags(Eq2())
-
-    assert hash(eq1) == hash(eq2) == hash(eq12)
-    assert hash(eq1) != hash(eq3)
-
-
 def test_strtobool():
     from pytools import strtobool
     assert strtobool("true") is True
@@ -736,8 +715,8 @@ def test_strtobool():
     assert strtobool("off") is False
 
     with pytest.raises(ValueError):
-        strtobool("tru")
-        strtobool("fal")
+        strtobool("tru")  # spellchecker: disable-line
+        strtobool("fal")  # spellchecker: disable-line
         strtobool("xxx")
         strtobool(".")
 
@@ -794,7 +773,7 @@ def test_typedump():
 
 
 def test_unique():
-    from pytools import unique
+    from pytools import unique, unique_difference, unique_intersection, unique_union
 
     assert list(unique([1, 2, 1])) == [1, 2]
     assert tuple(unique((1, 2, 1))) == (1, 2)
@@ -802,14 +781,27 @@ def test_unique():
     assert list(range(1000)) == list(unique(range(1000)))
     assert list(unique(list(range(1000)) + list(range(1000)))) == list(range(1000))
 
-    assert next(unique([1, 2, 1, 3])) == 1
-    assert next(unique([]), None) is None
-
     # Also test strings since their ordering would be thrown off by
     # set-based 'unique' implementations.
-    assert list(unique(["A", "B", "A"])) == ["A", "B"]
-    assert tuple(unique(("A", "B", "A"))) == ("A", "B")
-    assert next(unique(["A", "B", "A", "C"])) == "A"
+    assert list(unique(["a", "b", "a"])) == ["a", "b"]
+    assert tuple(unique(("a", "b", "a"))) == ("a", "b")
+
+    assert list(unique_difference(["a", "b", "c"], ["b", "c", "d"])) == ["a"]
+    assert list(unique_difference(["a", "b", "c"], ["a", "b", "c", "d"])) == []
+    assert list(unique_difference(["a", "b", "c"], ["a"], ["b"], ["c"])) == []
+
+    assert list(unique_intersection(["a", "b", "a"], ["b", "c", "a"])) == ["a", "b"]
+    assert list(unique_intersection(["a", "b", "a"], ["d", "c", "e"])) == []
+
+    assert list(unique_union(["a", "b", "a"], ["b", "c", "b"])) == ["a", "b", "c"]
+    assert list(unique_union(
+        ["a", "b", "a"], ["b", "c", "b"], ["c", "d", "c"])) == ["a", "b", "c", "d"]
+    assert list(unique(["a", "b", "a"])) == \
+        list(unique_union(["a", "b", "a"])) == ["a", "b"]
+
+    assert list(unique_intersection()) == []
+    assert list(unique_difference()) == []
+    assert list(unique_union()) == []
 
 
 # This class must be defined globally to be picklable
@@ -840,7 +832,7 @@ def test_record():
     assert str(r) == "SimpleRecord(a=1, b=2, c=3, d=4, e=5)"
 
     with pytest.raises(AttributeError):
-        r.ff
+        r.ff  # noqa: B018
 
     # Test pickling
     import pickle
@@ -884,6 +876,23 @@ def test_record():
     # }}}
 
 
+def test_permutations():
+    from math import factorial
+
+    from pytools import generate_permutations, generate_unique_permutations
+
+    perm = list(generate_permutations([1, 2, 3, 4]))
+    assert len(perm)  == factorial(4)
+    seq: Sequence[int] = (1, 3, 3, 4)
+    perm = list(generate_unique_permutations(seq))
+    assert len(perm)  == 12
+
+    perms = list(generate_permutations("1234"))
+    assert len(perms)  == factorial(4)
+    perms = list(generate_unique_permutations("1334"))
+    assert len(perms)  == 12
+
+
 if __name__ == "__main__":
     if len(sys.argv) > 1:
         exec(sys.argv[1])
